@@ -29,6 +29,10 @@ from chainlit.types import (
 )
 from chainlit.step import StepDict
 from chainlit.element import ElementDict
+from datetime import datetime, timezone
+
+def utc_now():
+    return datetime.now(timezone.utc).isoformat()
 
 AGENT_END_TOOL = "xpfinish-agent-execution-finished"
 MAX_TOOL_CALLS = 15
@@ -519,14 +523,15 @@ class ChainlitAdapter(SDKAdapter):
         
         return tool_calls
 
-    @cl.step(type="tool", name="xpander-ai")
-    async def process_tool_calls(self, tool_calls: Dict[int, ChainlitToolCallWithLLMResponse], on_node_enter: Optional[Callable], on_node_error: Optional[Callable], agent_custom_functions: list[GraphItem] = [], local_tools: Dict[str, Callable] = None) -> str | None:
+    async def process_tool_calls(self, run_id: str, tool_calls: Dict[int, ChainlitToolCallWithLLMResponse], on_node_enter: Optional[Callable], on_node_error: Optional[Callable], agent_custom_functions: list[GraphItem] = [], local_tools: Dict[str, Callable] = None) -> str | None:
         """
         Process tool calls by formatting their payloads and executing them.
 
         Args:
             tool_calls (Dict[int, ChainlitToolCallWithLLMResponse]): The tool calls to process.
         """
+        xpander_step = cl.Step(name="xpander-ai", type="tool", parent_id=run_id)
+        await xpander_step.send()
         
         current_execution = self.agent.execution.to_dict()
         user_settings = cl.user_session.get("user_settings",{})
@@ -579,11 +584,17 @@ class ChainlitAdapter(SDKAdapter):
         def get_custom_function_graph_item(tc: ToolCall) -> GraphItem | None:
             return next((gi for gi in agent_custom_functions if gi.name == tc.name or gi.name.replace(" ","_") == tc.name),None)
         
+        await xpander_step.remove()
         for raw_tool_call in tool_calls.values():
             tool_call = raw_tool_call.tool_call
             is_agent_end_tool = True if tool_call.name == AGENT_END_TOOL else False
             
-            current_step = cl.context.current_step
+            current_step = cl.Step(type="tool", parent_id=run_id)
+            
+            # current_step.parent_id = parent_message_id
+            current_step.fail_on_persist_error = True # stop in case of repeating error
+            current_step.start = current_step.created_at = utc_now()
+            
             if not is_debug_mode_active:
                 current_step.name = "xpander-ai"
                 current_step.show_input = False
@@ -591,6 +602,8 @@ class ChainlitAdapter(SDKAdapter):
             graph_item = self.agent.graph.find_node_by_item_id(item_id=tool_call.name) or self.agent.graph.find_node_by_name(name=tool_call.name) or get_custom_function_graph_item(tc=tool_call)
             tool_name = graph_item.name if graph_item else tool_call.name
             
+            is_sub_agent_root = False
+            should_send_step = True
             
             tool_agent_id = get_sub_agent_id_from_oas_by_name(agent_name=tool_call.name,oas=self.agent.oas)
             if tool_agent_id:
@@ -599,6 +612,8 @@ class ChainlitAdapter(SDKAdapter):
                     if on_node_enter:
                         await on_node_enter(tool_call=tool_call,graph_item=sub_agent_node)
                     tool_name = sub_agent_node.name
+                    current_step.id = run_id+"_"+sub_agent_node.item_id
+                    is_sub_agent_root = True
             else:
                 # report to parent window
                 graph_item = self.agent.graph.find_node_by_item_id(item_id=tool_call.name) or self.agent.graph.find_node_by_name(name=tool_call.name) or get_custom_function_graph_item(tc=tool_call)
@@ -609,8 +624,9 @@ class ChainlitAdapter(SDKAdapter):
                 if is_agent_end_tool:
                     current_step.default_open = True
                     tool_name = f"{self.agent.name} - Final Thoughts"
+                    should_send_step = False
                 else:
-                    tool_name = f"{self.agent.name} - {graph_item.name}"
+                    tool_name = graph_item.name
                     
             elif is_agent_end_tool:
                 tool_name = f"Crafting final answer"
@@ -619,6 +635,18 @@ class ChainlitAdapter(SDKAdapter):
                 current_step.name = tool_name
                 current_step.input = tool_call.payload
             
+            if is_sub_agent_root:
+                current_step.show_input = False
+                current_step.output = None
+            
+            
+            if is_sub_agent:
+                if not is_sub_agent_root:
+                    current_step.parent_id = run_id+"_"+self.agent.id
+            # send the step to the UI
+            if should_send_step:
+                await current_step.send()
+            
             # handle local tools
             if tool_call.type == ToolCallType.LOCAL:
                 tool_name = tool_call.name.replace("xpLocal_","")
@@ -644,6 +672,7 @@ class ChainlitAdapter(SDKAdapter):
                 
                 
                 # report result to the memory
+                await current_step.update()
                 self.agent.add_tool_call_results(tool_call_results=[tool_call_result])
             else:
                 # failures
@@ -664,14 +693,17 @@ class ChainlitAdapter(SDKAdapter):
                         ]
                     )
                     
-                    cl.context.current_step.show_input = False
-                    cl.context.current_step.default_open = True
-                    cl.context.current_step.output = error
-                    await cl.context.current_step.update()
+                    current_step.show_input = False
+                    current_step.default_open = True
+                    current_step.output = error
+                    if should_send_step:
+                        await current_step.update()
                     raise Exception(error)
                 else: # execute the tool
                     tool_call_result = self.agent.run_tool(tool=tool_call, payload_extension={"headers":{"x-xpander-source-node-type": SourceNodeType.ASSISTANT.value.lower()}})
             
+            current_step.end = utc_now()
+            
             if not tool_call_result.is_success and is_debug_mode_active:
                 current_step.is_error = True
                 await on_node_error(tool_call_result=tool_call_result,graph_item=graph_item)
@@ -681,7 +713,9 @@ class ChainlitAdapter(SDKAdapter):
             else:
                 self.execution_tool_calls[execution_id].append(tool_call_result.function_name)
             
-            if is_debug_mode_active:
+            if should_send_step:
+                await current_step.update()
+            if is_debug_mode_active and not is_sub_agent_root:
                 if is_agent_end_tool:
                     current_step.output = tool_call_result.payload['bodyParams']['result']
                     current_step.language = "text"
@@ -692,8 +726,12 @@ class ChainlitAdapter(SDKAdapter):
                 if tool_call_result.result == "STOP":
                         current_step.name = f"Waiting for human approval ({tool_name})"
                         current_step.output = None
+                
+                if should_send_step:
+                    await current_step.update()
                 if is_agent_end_tool and not is_sub_agent:
-                    await current_step.remove()
+                    if should_send_step:
+                        await current_step.remove()
                     tool_calls.clear()
                     return tool_call_result.payload['bodyParams']['result']
 
